use std::rc::Rc;

use anyhow::Error;
use itertools::Itertools;
use wasm_bindgen::{closure::Closure, prelude::wasm_bindgen, JsCast};
use wasm_bindgen_futures::js_sys::{Function, Object, Reflect};
use web_sys::{window, HtmlElement};
use yew::{
    function_component, html, use_effect_with, use_memo, use_node_ref, use_state, AttrValue, Html,
};

use super::{load_file_contents, run_script, EditorViewDetailProps};

#[wasm_bindgen(getter_with_clone)]
struct EditorViewConfig {
    #[allow(unused)]
    pub doc: String,
}

#[wasm_bindgen(getter_with_clone)]
struct EditorStateConfig {
    #[allow(unused)]
    pub doc: String,
    #[allow(unused)]
    pub extensions: Vec<CMExtension>,
}

#[wasm_bindgen(getter_with_clone)]
struct HighlightingConfig {
    #[allow(unused)]
    pub fallback: bool,
}

pub struct EditorKeepClosures {
    change_closure: Rc<Closure<dyn FnMut(ViewUpdate)>>,
}

#[wasm_bindgen(module = "@codemirror/state")]
extern "C" {
    type EditorState;
    #[wasm_bindgen(extends = Object)]
    type CMDocument;

    #[wasm_bindgen(static_method_of = EditorState)]
    fn create(settings: EditorStateConfig) -> EditorState;

    #[wasm_bindgen(method, getter)]
    fn doc(st: &EditorState) -> CMDocument;

    #[wasm_bindgen(js_namespace = ["EditorState", "allowMultipleSelections"], js_name = of)]
    fn editorstate_allow_multiple_selections_of(v: bool) -> CMExtension;
}

#[wasm_bindgen(getter_with_clone)]
struct CustomKeyBinding {
    pub key: String,
    pub mac: String,
    pub run: Function,
}

fn custom_key_binding(inp: CustomKeyBinding) -> KeyBinding {
    let kb = Object::new();
    let _ = Reflect::set(&kb, &"key".into(), &inp.key.into());
    let _ = Reflect::set(&kb, &"mac".into(), &inp.mac.into());
    let _ = Reflect::set(&kb, &"run".into(), &inp.run.into());
    kb.unchecked_into()
}

#[wasm_bindgen(module = "@codemirror/view")]
extern "C" {
    type EditorView;
    type ViewUpdate;
    #[derive(Clone)]
    type KeyBinding;
    #[derive(Clone)]
    type CMExtension;

    #[wasm_bindgen(constructor)]
    fn new(settings: EditorViewConfig) -> EditorView;

    #[derive(Clone)]
    type CMFacetUpdateListener;
    #[wasm_bindgen(js_namespace = ["EditorView"], js_name = updateListener, thread_local)]
    static UPDATE_LISTENER: CMFacetUpdateListener;

    #[wasm_bindgen(method)]
    fn of(facet: &CMFacetUpdateListener, cb: &Closure<dyn FnMut(ViewUpdate)>) -> CMExtension;

    #[wasm_bindgen(method, getter)]
    fn dom(upd: &EditorView) -> HtmlElement;

    #[wasm_bindgen(method, getter)]
    fn state(v: &EditorView) -> EditorState;

    #[wasm_bindgen(method, js_name = setState)]
    fn set_state(v: &EditorView, st: &EditorState);

    #[wasm_bindgen(method, getter, js_name = docChanged)]
    fn doc_changed(upd: &ViewUpdate) -> bool;

    #[wasm_bindgen(method, getter)]
    fn view(upd: &ViewUpdate) -> EditorView;

    #[wasm_bindgen(js_name = highlightSpecialChars)]
    fn highlight_special_chars() -> CMExtension;
    #[wasm_bindgen(js_name = drawSelection)]
    fn draw_selection() -> CMExtension;
    #[wasm_bindgen(js_name = highlightActiveLine)]
    fn highlight_active_line() -> CMExtension;
    #[wasm_bindgen(js_name = dropCursor)]
    fn drop_cursor() -> CMExtension;
    #[wasm_bindgen(js_name = rectangularSelection)]
    fn rectangular_selection() -> CMExtension;
    #[wasm_bindgen(js_name = crosshairCursor)]
    fn crosshair_cursor() -> CMExtension;
    #[wasm_bindgen(js_name = lineNumbers)]
    fn line_numbers() -> CMExtension;
    #[wasm_bindgen(js_name = highlightActiveLineGutter)]
    fn highlight_active_line_gutter() -> CMExtension;

    #[wasm_bindgen(js_namespace = keymap, js_name = of)]
    fn keymap_of(keys: Vec<KeyBinding>) -> CMExtension;
}

#[wasm_bindgen(module = "@codemirror/legacy-modes/mode/lua")]
extern "C" {
    #[wasm_bindgen(js_name = lua, thread_local)]
    static LUA: StreamLanguage;
}

#[wasm_bindgen(module = "@codemirror/language")]
extern "C" {
    #[derive(Clone)]
    type StreamLanguage;
    #[derive(Clone)]
    type CMHighlightStyle;

    #[wasm_bindgen(js_namespace = StreamLanguage, js_name = define)]
    fn streamlanguage_define(input: StreamLanguage) -> CMExtension;

    #[wasm_bindgen(js_name = defaultHighlightStyle, thread_local)]
    static DEFAULT_HIGHLIGHT_STYLE: CMHighlightStyle;

    #[wasm_bindgen(js_name = syntaxHighlighting)]
    fn syntax_highlighting(style: CMHighlightStyle, config: HighlightingConfig) -> CMExtension;

    #[wasm_bindgen(js_name = indentOnInput)]
    fn indent_on_input() -> CMExtension;
    #[wasm_bindgen(js_name = bracketMatching)]
    fn bracket_matching() -> CMExtension;
    #[wasm_bindgen(js_name = foldGutter)]
    fn fold_gutter() -> CMExtension;

    #[wasm_bindgen(js_name = foldKeymap, thread_local)]
    static FOLD_KEYMAP: Vec<KeyBinding>;
}

#[wasm_bindgen(module = "@codemirror/commands")]
extern "C" {
    #[wasm_bindgen(js_name = defaultKeymap, thread_local)]
    static DEFAULT_KEYMAP: Vec<KeyBinding>;
    #[wasm_bindgen(js_name = historyKeymap, thread_local)]
    static HISTORY_KEYMAP: Vec<KeyBinding>;

    #[wasm_bindgen]
    fn history() -> CMExtension;
}

#[wasm_bindgen(module = "@codemirror/autocomplete")]
extern "C" {
    #[wasm_bindgen]
    fn autocompletion() -> CMExtension;
    #[wasm_bindgen(js_name = closeBrackets)]
    fn close_brackets() -> CMExtension;

    #[wasm_bindgen(js_name = completionKeymap, thread_local)]
    static COMPLETION_KEYMAP: Vec<KeyBinding>;
    #[wasm_bindgen(js_name = closeBracketsKeymap, thread_local)]
    static CLOSE_BRACKETS_KEYMAP: Vec<KeyBinding>;
}

#[wasm_bindgen(module = "@codemirror/search")]
extern "C" {
    #[wasm_bindgen(js_name = searchKeymap, thread_local)]
    static SEARCH_KEYMAP: Vec<KeyBinding>;

    #[wasm_bindgen(js_name = highlightSelectionMatches)]
    fn highlight_selection_matches() -> CMExtension;
}

#[wasm_bindgen(module = "@codemirror/lint")]
extern "C" {
    #[wasm_bindgen(js_name = lintKeymap, thread_local)]
    static LINT_KEYMAP: Vec<KeyBinding>;
}

fn try_save_document(file: &str, contents: &str) -> anyhow::Result<()> {
    let win = window().ok_or_else(|| Error::msg("Can't get window"))?;
    let local = win
        .local_storage()
        .map_err(|_| Error::msg("Error retrieving localStorage"))?
        .ok_or_else(|| Error::msg("Local storage not available"))?;
    local
        .set(&format!("scriptfile_{}", file), contents)
        .map_err(|_| Error::msg("Error saving to localStorage"))
}

#[function_component(EditorArea)]
pub(super) fn editor_area(props: &EditorViewDetailProps) -> Html {
    let node_ref = use_node_ref();
    let editor_state = props.editor_state.clone();

    let closures = use_memo(editor_state.open_file.clone(), |_| EditorKeepClosures {
        change_closure: Closure::new(move |view_update: ViewUpdate| {
            if view_update.doc_changed() {
                if let Some(contents) = view_update.view().state().doc().to_string().as_string() {
                    match try_save_document(&editor_state.open_file, &contents) {
                        Ok(()) => {}
                        Err(e) => {
                            let mut new_state = (*editor_state.as_ref()).clone();
                            new_state.error_msg = Some(format!("Can't save: {}", e).into());
                            editor_state.set(new_state.into());
                        }
                    }
                }
            }
        })
        .into(),
    });
    let view = use_state(move || {
        EditorView::new(EditorViewConfig {
            doc: "Loading...".to_owned(),
        })
    });
    {
        let view = view.clone();

        let current_script = props.editor_state.open_file.clone();

        let set_err_state = props.editor_state.clone();
        let set_err = Rc::new(move |msg: Option<&str>| {
            let mut new_state = (*set_err_state.as_ref()).clone();
            new_state.error_msg = msg.map(|v| AttrValue::from(String::from(v)));
            set_err_state.set(new_state.into());
        });

        let global_memo_runkey = props.global_memo.clone();
        let frame_id = props.frame.clone();
        let runkey_closure = use_memo((), |()| {
            Closure::<dyn FnMut(EditorState) -> bool>::new(move |_ed_state| {
                run_script(
                    &current_script,
                    &global_memo_runkey,
                    set_err.clone(),
                    &frame_id,
                );
                true
            })
        });

        use_effect_with(props.editor_state.open_file.clone(), move |open_file| {
            let mut keymaps: Vec<KeyBinding> = vec![custom_key_binding(CustomKeyBinding {
                key: "Ctrl-Enter".to_owned(),
                mac: "Cmd-Enter".to_owned(),
                run: runkey_closure
                    .as_ref()
                    .as_ref()
                    .unchecked_ref::<Function>()
                    .clone(),
            })];
            keymaps.append(
                &mut vec![
                    &CLOSE_BRACKETS_KEYMAP,
                    &DEFAULT_KEYMAP,
                    &SEARCH_KEYMAP,
                    &HISTORY_KEYMAP,
                    &FOLD_KEYMAP,
                    &COMPLETION_KEYMAP,
                    &LINT_KEYMAP,
                ]
                .into_iter()
                .flat_map(|k| k.with(|v| v.clone()).into_iter())
                .collect_vec(),
            );

            view.set_state(&EditorState::create(EditorStateConfig {
                doc: load_file_contents(open_file)
                    .unwrap_or_else(|e| format!("Error loading content: {}", e.to_string())),
                extensions: vec![
                    line_numbers(),
                    highlight_active_line_gutter(),
                    highlight_special_chars(),
                    history(),
                    fold_gutter(),
                    draw_selection(),
                    drop_cursor(),
                    editorstate_allow_multiple_selections_of(true),
                    indent_on_input(),
                    syntax_highlighting(
                        DEFAULT_HIGHLIGHT_STYLE.with(CMHighlightStyle::clone),
                        HighlightingConfig { fallback: true },
                    ),
                    bracket_matching(),
                    close_brackets(),
                    autocompletion(),
                    rectangular_selection(),
                    crosshair_cursor(),
                    highlight_active_line(),
                    highlight_selection_matches(),
                    keymap_of(keymaps),
                    StreamLanguage::streamlanguage_define(LUA.with(StreamLanguage::clone)),
                    UPDATE_LISTENER
                        .with(CMFacetUpdateListener::clone)
                        .of(&closures.change_closure),
                ],
            }));
        });
    }
    use_effect_with(node_ref.clone(), move |node_ref| match node_ref.get() {
        None => {}
        Some(node) => {
            let _ = node.append_child(&view.dom());
        }
    });
    html! {
        <div class="d-flex w-100 h-100 flex-column">
          <div class="editorarea flex-grow-1 flex-shrink-1 d-flex flex-row" ref={node_ref.clone()}>
          </div>
        </div>
    }
}
