use std::collections::BTreeMap;

use crate::GlobalMemoCell;
use anyhow::Result;
use piccolo::{Callback, CallbackReturn, Context, FromValue, Function, Table, Value, Variadic};
use std::rc::Rc;
use wasm_bindgen::{prelude::Closure, JsCast, JsValue};
use web_sys::{console, XmlHttpRequest};

use super::named_closure::{release_named_global_closure, store_named_global_closure};

pub fn load_help(global_memo: &GlobalMemoCell, ctx: Context, url: &str) -> Result<()> {
    let xhr: Rc<XmlHttpRequest> = XmlHttpRequest::new()
        .map_err(|_e| anyhow::Error::msg("Error creating XmlHttpRequest"))?
        .into();
    let global_memo = global_memo.clone();
    let xhr_cb = xhr.clone();
    let load_callback: Closure<dyn Fn()> = Closure::new(move || {
        let _ = global_memo.lua_engine.borrow_mut().interp.try_enter(|ctx| {
            release_named_global_closure(ctx, "_help_load_callback");
            if let Ok(Some(txt)) = xhr_cb.response_text() {
                match serde_json::from_str::<BTreeMap<String, String>>(&txt) {
                    Err(_) => console::log_1(&JsValue::from_str(
                        "Loading help failed because result wasn't a String -> String map",
                    )),
                    Ok(d) => {
                        let help_tbl: Value = ctx.get_global("help")?;
                        let help_tbl = if help_tbl.is_nil() {
                            let tbl = Table::new(&ctx);
                            ctx.set_global("help", tbl);
                            tbl
                        } else {
                            Table::from_value(ctx, help_tbl)?
                        };
                        for (k, v) in d.iter() {
                            help_tbl.set(ctx, k.clone(), v.clone())?;
                        }
                    }
                }
            }
            Ok(())
        });
    });
    xhr.set_onload(Some(load_callback.as_ref().unchecked_ref()));
    store_named_global_closure(ctx, "_help_load_callback", load_callback);
    xhr.open("GET", url)
        .map_err(|_e| anyhow::Error::msg("Error opening XmlHttpRequest"))?;
    xhr.send()
        .map_err(|_e| anyhow::Error::msg("Error sending XmlHttpRequest"))?;

    Ok(())
}

pub(super) fn cmd_help<'gc>(ctx: Context<'gc>) -> Callback<'gc> {
    Callback::from_fn(&ctx, move |ctx, _ex, mut stack| {
        let help: Value = ctx.get_global("help")?;
        let echo: Function = ctx.get_global::<Table>("commands")?.get(ctx, "echo")?;
        let args: Variadic<Vec<String>> = stack.consume(ctx)?;
        let mut path: String = args.join(" ");
        if let Some(p) = path.strip_prefix('#') {
            path = p.to_owned();
        }
        path = path.trim().to_lowercase();

        if path.is_empty() {
            path = "_index".to_owned();
        }

        if help.is_nil() {
            stack.into_back(ctx, "Help has not (yet) loaded successfully");
            return Ok(CallbackReturn::Call {
                function: echo,
                then: None,
            });
        }
        let help: Table = Table::from_value(ctx, help)?;
        let help_value: Value = help.get(ctx, path)?;
        let help_value = if help_value.is_nil() {
            "No help available on that.".to_owned()
        } else {
            String::from_value(ctx, help_value)?
        };
        stack.into_back(ctx, help_value);

        Ok(CallbackReturn::Call {
            function: echo,
            then: None,
        })
    })
}
