use anyhow::Result;
use piccolo::{Callback, CallbackReturn, Context, IntoValue, Table};
use std::{ops::Deref, rc::Rc};
use yew::UseStateSetter;

use crate::{navbar::NavbarButton, FrameId, GlobalLayoutCell, GlobalLayoutState, GlobalMemoCell};

pub fn init_navbar(
    ctx: Context,
    global_memo: &GlobalMemoCell,
    global_layout: &UseStateSetter<GlobalLayoutCell>,
) -> Result<()> {
    let info_tbl: Table = ctx.get_global::<Table>("info")?;
    let nav_tbl: Table = Table::new(&ctx);

    let default_buttons: Vec<NavbarButton> = vec![
        NavbarButton {
            icon: "code-square".to_owned(),
            tooltip: "Open the script editor".to_owned(),
            frame: FrameId(1),
            command: "#editor".to_owned(),
        },
        NavbarButton {
            icon: "arrow-left".to_owned(),
            tooltip: "Go west".to_owned(),
            frame: FrameId(1),
            command: "w".to_owned(),
        },
        NavbarButton {
            icon: "arrow-up-left".to_owned(),
            tooltip: "Go northwest".to_owned(),
            frame: FrameId(1),
            command: "nw".to_owned(),
        },
        NavbarButton {
            icon: "arrow-up".to_owned(),
            tooltip: "Go north".to_owned(),
            frame: FrameId(1),
            command: "n".to_owned(),
        },
        NavbarButton {
            icon: "arrow-up-right".to_owned(),
            tooltip: "Go northeast".to_owned(),
            frame: FrameId(1),
            command: "ne".to_owned(),
        },
        NavbarButton {
            icon: "arrow-right".to_owned(),
            tooltip: "Go east".to_owned(),
            frame: FrameId(1),
            command: "e".to_owned(),
        },
        NavbarButton {
            icon: "arrow-down-right".to_owned(),
            tooltip: "Go southeast".to_owned(),
            frame: FrameId(1),
            command: "se".to_owned(),
        },
        NavbarButton {
            icon: "arrow-down".to_owned(),
            tooltip: "Go south".to_owned(),
            frame: FrameId(1),
            command: "s".to_owned(),
        },
        NavbarButton {
            icon: "arrow-down-left".to_owned(),
            tooltip: "Go southwest".to_owned(),
            frame: FrameId(1),
            command: "sw".to_owned(),
        },
    ];

    let mut i: i64 = 1;
    for button in &default_buttons {
        nav_tbl.set(ctx, i, button.clone().into_value(ctx))?;
        i += 1;
    }
    nav_tbl.set(ctx, "showing", true)?;
    info_tbl.set(ctx, "navbar", nav_tbl)?;

    let mut new_layout: GlobalLayoutState = global_memo.layout.borrow().deref().deref().clone();
    new_layout.navstate.buttons = default_buttons;
    let new_layout: Rc<GlobalLayoutState> = new_layout.into();
    *global_memo.layout.borrow_mut() = new_layout.clone();
    global_layout.set(new_layout);
    Ok(())
}

pub fn syncnavbar<'gc>(
    ctx: Context<'gc>,
    global_memo: &GlobalMemoCell,
    global_layout: &UseStateSetter<GlobalLayoutCell>,
) -> Callback<'gc> {
    let global_memo = global_memo.clone();
    let global_layout = global_layout.clone();
    Callback::from_fn(&ctx, move |ctx, _ex, _stack| {
        let mut new_layout: GlobalLayoutState = global_memo.layout.borrow().deref().deref().clone();

        let tbl: Table = ctx.get_global::<Table>("info")?.get(ctx, "navbar")?;

        new_layout.navstate.showing = tbl.get(ctx, "showing")?;
        let mut buttons: Vec<NavbarButton> = vec![];
        let mut i: i64 = 1;
        loop {
            match tbl.get(ctx, i) {
                Err(_) => break,
                Ok(v) => buttons.push(v),
            }
            i += 1;
        }
        new_layout.navstate.buttons = buttons;
        let new_layout: Rc<GlobalLayoutState> = new_layout.into();
        *global_memo.layout.borrow_mut() = new_layout.clone();
        global_layout.set(new_layout);
        Ok(CallbackReturn::Return)
    })
}
