use piccolo::{Context, FromValue, IntoValue, Table, Value};
use yew::{function_component, html, Callback, Html, Properties, UseStateHandle};

use crate::{command_handler::command_handler, FrameId, GlobalLayoutCell, GlobalMemoCell};

#[derive(Properties, PartialEq)]
pub struct NavbarProps {
    pub global_memo: GlobalMemoCell,
    pub global_layout: UseStateHandle<GlobalLayoutCell>,
}

#[derive(PartialEq, Clone, Debug)]
pub struct NavbarButton {
    pub icon: String,
    pub tooltip: String,
    pub frame: FrameId,
    pub command: String,
}

impl<'gc> IntoValue<'gc> for NavbarButton {
    fn into_value(self, ctx: Context<'gc>) -> Value<'gc> {
        let tbl = Table::new(&ctx);
        let _ = tbl.set(ctx, "icon", self.icon);
        let _ = tbl.set(ctx, "tooltip", self.tooltip);
        let _ = tbl.set(ctx, "frame", self.frame.0 as i64);
        let _ = tbl.set(ctx, "command", self.command);
        tbl.into_value(ctx)
    }
}
impl<'gc> FromValue<'gc> for NavbarButton {
    fn from_value(ctx: Context<'gc>, value: Value<'gc>) -> Result<Self, piccolo::TypeError> {
        let tbl: Table = Table::from_value(ctx, value)?;
        Ok(NavbarButton {
            icon: tbl.get(ctx, "icon")?,
            tooltip: tbl.get(ctx, "tooltip")?,
            frame: FrameId(tbl.get(ctx, "frame")?),
            command: tbl.get(ctx, "command")?,
        })
    }
}

#[derive(PartialEq, Clone)]
pub struct NavState {
    pub showing: bool,
    pub buttons: Vec<NavbarButton>,
}
impl Default for NavState {
    fn default() -> Self {
        Self {
            showing: true,
            // Overridden later in init_navbar, starts off minimal.
            buttons: vec![],
        }
    }
}

#[function_component(Navbar)]
pub fn navbar(props: &NavbarProps) -> Html {
    if !props.global_layout.navstate.showing {
        return html! { <></> };
    }
    html! {
        <nav class="navbar navbar-expand-lg bg-body-tertiary">
            <div>
        {
            props.global_layout.navstate.buttons.iter().map(|button| {
                let global_memo = props.global_memo.clone();
                let button = button.clone();
                let onclick = Callback::from(move |_| {
                    command_handler(&global_memo, &button.frame, &button.command);
                });
                html! {
                    <button type="button" title={button.tooltip} class="btn" {onclick}>
                        <i class={format!("bi bi-{}", button.icon)}></i>
                    </button>
                }
            }).collect::<Vec<_>>()
        }
            </div>
        </nav>
    }
}

#[cfg(test)]
mod test {
    use piccolo::{FromValue, IntoValue, Lua};

    use crate::FrameId;

    use super::NavbarButton;

    #[test]
    fn button_tovalue_fromvalue_roundtrips() {
        let mut lua = Lua::empty();
        lua.enter(|ctx| {
            let b = NavbarButton {
                icon: "myicon".to_owned(),
                tooltip: "I'm a tooltip".to_owned(),
                frame: FrameId(42),
                command: "hello world".to_owned(),
            };
            assert_eq!(
                NavbarButton::from_value(ctx, b.clone().into_value(ctx)).unwrap(),
                b
            );
        })
    }
}
